


// optimierung: in file.c die variable seek registrieren in register


/*
*******************************************************************************************************************************
CHANGELOG:

  mmc-0.4.5
  struct file -> "unsigned char name" wird zu "char name"

*******************************************************************************************************************************

EXTERN genutzte funktionen ! siehe fat-0.x.h
extern unsigned char 	 	fat_initfat(void);							// initalisierung (durchsucht MBR oder nicht) 
extern unsigned long int 	fat_clustToSec(unsigned long int);			// rechnet cluster zu 1. sektor des clusters um
extern unsigned char 		fat_writeSector(unsigned long int sec);		// schreibt sektor auf karte
extern unsigned long int 	fat_getNextCluster(unsigned long int oneCluster);	// fat auf nächsten, verketteten cluster durchsuchen 
extern unsigned long int 	fat_getFreeCluster(unsigned long int offsetCluster);	// sucht leeren fat cluster aus der fat
extern unsigned char		fat_setCluster(unsigned long int cluster, unsigned long int content); // setzt cluster inhalt in der fat
extern unsigned char 		fat_delClusterChain(unsigned long int startCluster);	// löscht cluster-chain in der fat
extern unsigned char 		fat_getFreeRowOfDir(unsigned long int dir);				// durchsucht directory nach feiem eintrag
extern unsigned char 		fat_makeFileEntry(unsigned char name[],unsigned char attrib,unsigned long int cluster,unsigned long int length);
extern unsigned char 		fat_loadSector(unsigned long int sec);				// läd übergebenen absoluten sektor
extern unsigned char 		fat_loadFileDataFromDir(unsigned char name[]);		// durchsucht das aktuelle directory 
extern unsigned char		fat_cd(unsigned char *);							// wechselt directory (start im rootDir)
*/


//***************************************************************************************************************
// nur intern genutzte funktionen ! extern benutzte funktionen siehe : fat-0.x.h !

unsigned char 		fat_loadFatData(unsigned long int);					// läd fat daten 
void 				fat_markSector00(void);								// markiert puffer:sector mit 0x00

// fat_getFreeRowOfCluster -> fat_getFreeRowOfDir -> fat_makeRowDataEntry -> fat_makeFileEntry -> fat_writeSector  "eintrag gemacht !!"
unsigned char     	fat_getFreeRowOfCluster(unsigned long secStart);	// durchsucht cluster nach freiem eintrag
unsigned char 		fat_makeRowDataEntry(unsigned int row,unsigned char name[],unsigned char attrib,unsigned long int cluster,unsigned long int length);

// fat_loadSector -> fat_loadRowOfSector -> fat_loadFileDataFromCluster -> fat_loadFileDataFromDir -> fat_cd   "daten chain"
unsigned char 		fat_loadRowOfSector(unsigned int);					// läd reihe des geladen sektors auf struct:file
unsigned char 		fat_loadFileDataFromCluster(unsigned long int sec , unsigned char name[]);	// durchsucht die reihen des geladenen sektors 

//***************************************************************************************************************

unsigned char 	sector[512];				// sektor puffer
unsigned char	*bytesOfSec=&sector[0];		// byte zeiger auf puffer:sector
unsigned char 	bufferDirty=0;				// zeigt an ob buffer dirty (wird von fat_setCluster gesetzt und von fat_loadSector gelesen)


struct Fat{									// fat daten (1.cluster, root-dir, dir usw.)
	unsigned char 		secPerClust;		// anzahl der sektoren pro cluster
	unsigned long int 	currentSectorNr; 	// aktuell geladener Sektor (in sector)  //beschleunigt wenn z.b 2* 512 byte puffer vorhanden, oder bei fat operationen im gleichen sektor		
	unsigned long int 	dir; 		  		// Direktory zeiger rootDir=='0' sonst(1.Cluster des dir; start auf root)	
	unsigned long int  	rootDir;			// Sektor(f16)/Cluster(f32) nr root directory
	unsigned long int  	dataDirSec;			// Sektor nr data area 
	unsigned long int  	fatSec;	 			// Sektor nr fat area
	unsigned int  		bytsPerSec;			// bytes per sector	
	unsigned char  		fatType;			// fat16 oder fat32 (16 oder 32)
	
	} fat;


struct File{								// datei infos	(32 bytes lang- 16 bei 512 bytesPerSec)												
	char		 		name[13];			// 0,10			datei Name.ext (8.3 = max 11)
	unsigned char		attrib;				// 11,1			datei Attribut: 8=value name, 32=datei, 16=Verzeichniss, 15=linux kleingeschrieben eintrag
	unsigned long int 	firstCluster;		// 20,2 /26,2	datei 1.cluster hi,low(möglicherweise der einzige)	(4-byte)
	unsigned long int	length;				// 28,4			datei Länge (4-byte)				
	unsigned char 		row;				// reihe im sektor in der die datei infos stehen (reihe 0-15)

	} file;	


#if (write==0)
//***************************************************************************************************************
// schreibt sektor nummer:sec auf die karte (puffer:sector) !!
// setzt bufferFlag=0 da puffer nicht dirty sein kann nach schreiben !
//***************************************************************************************************************
unsigned char fat_writeSector(unsigned long int sec){	

  if(0==mmc_write_sector(sec,sector)){
	bufferDirty=0;	// buffer kann nicht dirty sein weil grade geschrieben		
	return(0);		// schreiben von sektor puffer											
	}

  return (1);
}

//***************************************************************************************************************
// markiert sector komplett mit 0x00 (nur den gepufferten)
// wenn neuer direktory cluster verkettet wird, muss er mit 0x00 initialisiert werden (alle felder)
//***************************************************************************************************************
void fat_markSector00(void){
		
  fat.currentSectorNr=0;	// geladener sektor wird hier geändert !
  unsigned int i=511;	
	
  do{
	sector[i]=0x00;		//schreibt puffer sector voll mit 0x00==leer
	}while(--i);		
  sector[0]=0x00;		// noch die stelle 0 überschreiben		
}

#endif



// ***************************************************************************************************************
// umrechnung cluster auf 1.sektor des clusters (möglicherweise mehrere sektoren/cluster !
// ***************************************************************************************************************
unsigned long int fat_clustToSec(unsigned long int clust){

	return fat.dataDirSec+2+((clust-2) * fat.secPerClust);		// errechnet den 1. sektor der sektoren des clusters	
}



//***************************************************************************************************************
// läd sektor:sec auf puffer:sector zum bearbeiten im ram !
// setzt currentSectorNr auf richtigen wert (also den sektor der gepuffert ist). es wird geprüft
// ob der gepufferte sektor geändert wurde, wenn ja muss erst geschrieben werden, um diese daten nicht zu verlieren !
//***************************************************************************************************************
unsigned char fat_loadSector(unsigned long int sec){
	
  if(sec!=fat.currentSectorNr){					// nachladen nötig		
	#if (write==0)
	if(bufferDirty==1)fat_writeSector(fat.currentSectorNr);	// puffer diry, also vorher schreiben		  		  
	#endif
	if(0==mmc_read_sector(sec,sector)){			// neuen sektor laden					
	  fat.currentSectorNr=sec;					// aktualisiert sektor nummer (nummer des gepufferten sektors)
	  return(0);
	  }
	}	
		
  else return(0);								// alles ok, daten sind schon da (sec==fat.currentSectorNr)	

  return(1);									// fehler	
}







// datei lesen funktionen:

// fat_loadSector -> fat_loadRowOfSector -> fat_loadFileDataFromCluster -> fat_loadFileDataFromDir -> fat_loadFileDataFromDir -> fat_cd   "daten chain"

//***************************************************************************************************************
// läd die reihe:row des gepufferten sektors auf das struct:file. dort stehen dann
// alle wichgigen daten wie: 1.cluster,länge bei dateien, name des eintrags, reihen nummer (im sektor), attribut use...
//***************************************************************************************************************
unsigned char fat_loadRowOfSector(unsigned int row){

	unsigned char i;						// zähler im sektor(32byte zähler).
	row=row<<5;								// multipliziert mit 32 um immer auf zeilen anfang zu kommen (zeile 0=0,zeile 1=32,zeile 2=62 usw)	.

	bytesOfSec=&sector[row];				// byte zeiger auf puffer:sector.
	
	void *firstCluster=&file.firstCluster;	// void pointer auf datei erster cluster,zum schreiben von einzel bytes auf 4 byte variable.
	void *length=&file.length; 				// void pointer auf datei länge, zum schreiben von einzel bytes auf 4 byte variable.

	for(i=0;i<11;i++)  file.name[i]=*bytesOfSec++;		// datei name, ersten 10 bytes vom 32 byte eintrag.		  

	file.attrib=*bytesOfSec;							// datei attribut, byte 11.
	
	*(unsigned char*)firstCluster++=sector[row+26];		// datei erster cluster	, byte von klein nach hoch: 26,27,20,21	.
	*(unsigned char*)firstCluster++=sector[row+27];		// hier nicht mit bytesOfSec pointer, weil man hin und her springen,
	*(unsigned char*)firstCluster++=sector[row+20];		// müsste..
	*(unsigned char*)firstCluster++=sector[row+21];

	bytesOfSec+=17;

	*(unsigned char*)length++=*bytesOfSec++;			// datei länge, bytes 28,29,30,31.
	*(unsigned char*)length++=*bytesOfSec++;
	*(unsigned char*)length++=*bytesOfSec++;
	*(unsigned char*)length++=*bytesOfSec++;

	return(0);				
}


//***************************************************************************************************************
// geht reihen weise durch sektoren des clusters mit dem startsektor:sec, und sucht nach der datei mit dem 
// namen:name. es werden die einzelnen sektoren nachgeladen auf puffer:sector vor dem bearbeiten.
// wird die datei in dem cluster gefunden ist return 0 , sonst return1.
//***************************************************************************************************************
unsigned char fat_loadFileDataFromCluster(unsigned long int sec , unsigned char name[]){
	
  unsigned char r;  
  unsigned char s=0;

  do{											// sektoren des clusters prüfen
	if(0==fat_loadSector(sec+s)){				// läd den sektor:sec auf den puffer:sector		
	  for(r=0;r<16;r++){						// zählt zeilennummer (16(zeilen) * 32(spalten) == 512 bytes des sektors)
		fat_loadRowOfSector(r);					// zeile 0-15 auf struct:file laden 
		if(0==strncmp(file.name,(char*)name,10)){	// zeile r ist gesuchte
		  file.row=r;										// zeile sichern.
		  return(0);			
		  }
		}
	}
	s++;
	}while(s<fat.secPerClust);		// geht durch sektoren des clusters

	return (1);									// fehler (datei nicht gefunden, oder fehler beim lesen)
}


//***************************************************************************************************************
// wenn dir == 0 dann wird das root direktory durchsucht, wenn nicht wird der ordner cluster-chain gefolgt, um
// die datei zu finden. es wird das komplette directory in dem man sich befindet durchsucht.
// bei fat16 wird der rootDir berreich durchsucht, bei fat32 die cluster chain des rootDir.
//***************************************************************************************************************
unsigned char fat_loadFileDataFromDir(unsigned char name[]){

  unsigned int s;  

  if(fat.dir==0 && fat.fatType==16){									// IM ROOTDIR.	fat16		
	for(s=0;s<(unsigned int)(fat.dataDirSec+2-fat.rootDir);s++){			// zählt durch RootDir sektoren (errechnet anzahl rootDir sektoren).								
	  if(0==fat_loadFileDataFromCluster(fat.rootDir+s,name))return(0);		// sucht die datei, wenn da, läd daten (1.cluster usw)															
	  }
	}		
	
  else {		
	unsigned long int i;
	if(fat.dir==0 && fat.fatType==32)i=fat.rootDir;						// IM ROOTDIR.	fat32		
	else i=fat.dir;																												
	while(!((i==0xfffffff&&fat.fatType==32)||(i==0xffff&&fat.fatType==16))){	// prüft ob weitere sektoren zum lesen da sind (fat32||fat16)					=
	  if(0==fat_loadFileDataFromCluster(fat_clustToSec(i),name))return(0);		// lät die daten der datei auf struct:file. datei gefunden (umrechnung auf absoluten sektor)																									
	  i=fat_getNextCluster(i);													// liest nächsten cluster des dir-eintrags (unterverzeichniss größer 16 einträge)								
	  }
	}

  return(1);																	// datei/verzeichniss nicht gefunden
}



//***************************************************************************************************************
// start immer im root Dir. start in root dir (dir==0).
// es MUSS in das direktory gewechselt werden, in dem die datei zum lesen/anhängen ist (außer root, da startet mann)!
//***************************************************************************************************************
unsigned char fat_cd(unsigned char name[]){

  if(name[0]==0){								// ZUM ROOTDIR FAT16/32
	fat.dir=0;										// root dir 
	return(0);
	}

  if(0==fat_loadFileDataFromDir(name)){			// NICHT ROOTDIR	(fat16/32)
	fat.dir=file.firstCluster;						// zeigt auf 1.cluster des dir	(fat16/32)	
	return(0);
	}

  return(1);									// dir nicht gewechselt (nicht da?!?) !!
}






#if (write==0)
// datei anlegen funktionen :

//fat_getFreeRowOfCluster -> fat_getFreeRowOfDir -> fat_makeRowDataEntry -> fat_makeFileEntry -> fat_writeSector -> eintrag gemacht !!

// ***************************************************************************************************************
// sucht leeren eintrag (zeile) im cluster mit dem startsektor:secStart.
// wird dort kein freier eintrag gefunden ist return (1).
// wird ein freier eintrag gefunden, ist die position der freien reihe auf file.row abzulesen und return (0).
// der sektor mit der freien reihe ist auf dem puffer:sector gepuffert.
// ****************************************************************************************************************
unsigned char fat_getFreeRowOfCluster(unsigned long secStart){
  
  unsigned int 	i;										// 1. zeichen einer reihe, eine sektors, eines clusters.
  unsigned char s=0;									// sektoren des clusters.
  
  do{
	file.row=0;											// neuer sektor(oder 1.sektor), reihen von vorne.
	if(0==fat_loadSector(secStart+s)){					// läd sektor auf puffer:buffer.dSector, setzt buffer.currentDatSector.
	  for(i=0;i<512;i=i+32){							// zählt durch zeilen (0-15).									  
		if(sector[i]==0x00||sector[i]==0xE5)return(0);	// prüft auf freihen eintrag (leer oder gelöscht == OK!).
		file.row++;										// zählt reihe hoch (nächste reihe im sektor).	
		}
	  }	
	s++;												// sektoren des clusters ++ weil einen geprüft.	
	}while(s<fat.secPerClust);							// geht die sektoren des clusters durch (möglicherweise auch nur 1. sektor).
  return (1);											// nicht gefunden in diesem cluster (== nicht OK!).
}


// ***************************************************************************************************************
// sucht leeren eintrag (zeile) im directory mit dem startcluster:dir.
// geht die cluster chain des direktories durch. dabei werden auch alle sektoren der cluster geprüft.
// wird dort kein freier eintrag gefunden, wird ein neuer leerer cluster gesucht, verkettet und der 
// 1. sektor des freien clusters geladen. die reihe wird auf den ersten eintrag gesetzt, da frei.
// anhand der reihe kann man nun den direktory eintrag vornehmen, und auf die karte schreiben.
// ****************************************************************************************************************
unsigned char fat_getFreeRowOfDir(unsigned long int dir){
  
  unsigned long int start=dir;   
 
  // solange bis ende cluster chain.
  while( !((dir==0xfffffff&&fat.fatType==32)||(dir==0xffff&&fat.fatType==16)) ){		  
	if(0==fat_getFreeRowOfCluster(fat_clustToSec(dir)))return(0);	// freien eintrag in clustern, des dir gefunden !!
	start=dir;		
	dir=fat_getNextCluster(dir);	
	}										// wenn aus schleife raus, kein freier eintrag da -> neuer cluster nötig.

  dir=fat_getFreeCluster(start);			// muss neuen freien cluster suchen 	(benutzt puffer:sector).  
  fat_setCluster(start,dir);				// cluster-chain mit neuem verketten	(benutzt puffer:sector).	
  fat_setCluster(dir,0x0fffffff);			// cluster-chain ende markieren	  		(benutzt puffer:sector).    
  fat_loadSector(fat_clustToSec(dir));		// läd neuen cluster (wenn bufferFlag==1 schreibt vorher alten cluster) !!
  fat_markSector00();						// löschen des sektors (nur im puffer) (benutzt puffer:sector).

  unsigned char i;
  for(i=1;i<fat.secPerClust;i++){			// i=1 weil der 1. eintrag eh noch geschrieben wird... wäre hier doppelt.
	fat_writeSector(fat.currentSectorNr+i);	// löschen des cluster (überschreibt mit 0x00), wichtig bei ffls,
	}										// weil wenn cluster nicht mit 0x00 überschrieben wird, werden mögilcherwies
											// alte daten ausgegeben !!!!
  file.row=0;								// erste reihe frei, weil grad neuen cluster verkettet.

  return(0);						// keinen freien platz in clustern, des dir gefunden, aber neuen cluster verkettet.
}


//***************************************************************************************************************
// erstellt 32 byte eintrag einer datei, oder verzeichnisses im puffer:sector.
// erstellt eintrag in reihe:row, mit namen:name usw... !!  
// muss noch auf die karte geschrieben werden ! nicht optimiert auf geschwindigkeit.
//***************************************************************************************************************
unsigned char fat_makeRowDataEntry(unsigned int row,unsigned char name[],unsigned char attrib,unsigned long int cluster,unsigned long int length){

  bufferDirty=1;								// puffer beschrieben, also neue daten darin(vor lesen muss geschrieben werden)

  unsigned char i;								//byte zähler in reihe von sektor (32byte eintrag)
  row=row<<5;									//multipliziert mit 32 um immer auf zeilen anfang zu kommen (zeile 0=0,zeile 1=32,zeile 2=62 ... zeile 15=480)	

  bytesOfSec=&sector[row];						// init des zeigers
				
  for(i=0;i<32;i++){													

  if(i<11)*bytesOfSec=name[i];					//namen schreiben (format beachten!!) (10 byte)

	if(i==11)*bytesOfSec=attrib;				//attrib schreiben (1 byte)

	if(i>11&&i<20)*bytesOfSec=0x01;				//nulle nicht nötige felder !
	if(i>21&&i<26)*bytesOfSec=0x01;				//nulle nicht nötige felder !

	if(i==20){									// 1. cluster der datei byte 20/21 und 26/27 = 4 bytes
	  *bytesOfSec++=(cluster&0x00ff0000)>>16;	// 1. low							
	  *bytesOfSec=(cluster&0xff000000)>>24;		// high byte																
	  i++;
	  }	
		
	if(i==26){												
	  *bytesOfSec++=(cluster&0x000000ff);		// low byte														
	  *bytesOfSec=(cluster&0x0000ff00)>>8;		// 2. low						
	  i++;
	  }	
		
	if(i==28){									//datei länge (4 bytes)					
	  void *vLength=&length;
	  *bytesOfSec++ =*(unsigned char*)vLength++;	// low						
	  *bytesOfSec++ =*(unsigned char*)vLength++;	//	1. hi			
	  *bytesOfSec++ =*(unsigned char*)vLength++;	//	2. hi 				
	  *bytesOfSec =*(unsigned char*)vLength;		//	hi		
	  return (0);								// alles ok, eintrag angelegt !!		
	  }
	bytesOfSec++;		
	}
		
  return (1);		//fehler (datei nicht gefunden, oder fehler beim lesen)
}


//***************************************************************************************************************
// macht den datei eintrag im jetzigen verzeichniss (fat.dir).
// file.row enthält die reihen nummer des leeren eintrags, der vorher gesucht wurde, auf puffer:sector ist der gewünschte
// sektor gepuffert. für fat16 im root dir muss andere funktion genutzt werden, als fat_getFreeRowOfDir (durchsucht nur dirs).
// fat.rootDir enthält bei fat32 den start cluster des directory, bei fat16 den 1. sektor des rootDir bereichs!
//***************************************************************************************************************
unsigned char fat_makeFileEntry(unsigned char name[],unsigned char attrib,unsigned long int cluster,unsigned long int length){
  
  unsigned int s;														// zähler für root dir sektoren fat16
 
  if(fat.dir==0&&fat.fatType==32)fat_getFreeRowOfDir(fat.rootDir);	// IM ROOT DIR (fat32)

  else if(fat.dir==0 && fat.fatType==16){							// IM ROOT DIR (fat16)		
	for(s=0;s<(unsigned int)(fat.dataDirSec+2-fat.rootDir);s++){		// zählt durch RootDir sektoren (errechnet anzahl rootDir sektoren).								
	  if(0==fat_getFreeRowOfCluster(fat.rootDir+s))break;				// geht durch sektoren des root dir.
	  }
	}  

  else fat_getFreeRowOfDir(fat.dir);								// NICHT ROOT DIR
		
  fat_makeRowDataEntry(file.row,name,attrib,cluster,length);			// macht file eintrag im puffer	
  fat_writeSector(fat.currentSectorNr);									// schreibt file daten auf karte		  
  
  return(0);
	  
}

#endif





// fat funktionen:

//***************************************************************************************************************
// sucht nötige folge Cluster aus der fat ! 
// erster daten cluster = 2, ende einer cluster chain 0xFFFF (fat16) oder 0xFFFFFFF, 0xFFFFFF8 (fat32), 
// stelle des clusters in der fat, hat als wert, den nächsten cluster. (1:1 gemapt)!
//***************************************************************************************************************
unsigned long int fat_getNextCluster(unsigned long int oneCluster){	

  // FAT 16**************FAT 16	
  if(fat.fatType==16){					
												// i=oneCluster/256
	unsigned long int i=oneCluster>>8;;			// errechnet den sektor der fat in dem oneCluster ist (rundet immer ab) 
												// j=(oneCluster-256*i)*2 == 2*oneCluster-512*i
	unsigned long int j=(oneCluster<<1)-(i<<9);	// errechnet das low byte von oneCluster in dem ausgelesenen sektor der fat (high und low byte enthalten den folge cluster)
			
	if(0==fat_loadSector(i+fat.fatSec)){		// ob neu laden nötig wird von fat_loadSector geprüft
	  i=0;  
	  void *vi=&i;								// zeiger auf i
	  bytesOfSec=&sector[j];	  				// zeiger auf puffer
	  *(unsigned char*)vi++=*bytesOfSec++;		// setzen low byte von i, aus puffer	  
	  *(unsigned char*)vi++=*bytesOfSec++;		// setzen von höherem byte in i, aus puffer
	  return i;									// gibt neuen cluster zurück (oder 0xffff)
	  }	
	}

  // FAT 32**************FAT 32	
  else{
												// i=oneCluster/128
	unsigned long int i=oneCluster>>7;			// errechnet den sektor der fat in dem oneCluster ist (rundet immer ab) 
												// j=(oneCluster-128*i)*4 == oneCluster*4-512*i
	unsigned long int j=(oneCluster<<2)-(i<<9);	// errechnet das low byte von oneCluster in dem ausgelesenen sektor der fat (high und low byte enthalten den folge cluster)		
		
	if(0==fat_loadSector(i+fat.fatSec)){		// ob neu laden nötig wird von fat_loadSector geprüft
	  void *vi=&i;
	  bytesOfSec=&sector[j];
	  *(unsigned char*)vi++=*bytesOfSec++;
	  *(unsigned char*)vi++=*bytesOfSec++;	 
	  *(unsigned char*)vi++=*bytesOfSec++;
	  *(unsigned char*)vi++=*bytesOfSec++;	 
	  return i;
	  }	
  }		

  return(0);									// neuladen des fat sektors, in dem oneCluster ist nötig !!		
}



#if (write==0)
//***************************************************************************************************************
// sucht freien cluster in der fat ab dem offsetCluster (also offsetCluster+1).
//***************************************************************************************************************
unsigned long int fat_getFreeCluster(unsigned long int offsetCluster){
	
  do{											
	offsetCluster++;							// cluster nummer	
	}while(fat_getNextCluster(offsetCluster));	// freier cluster gefunden (return 0, weil inhalt des clusters 0)

  return offsetCluster;							
}


//***************************************************************************************************************
// setzt den cluster inhalt. errechnet den sektor der fat in dem cluster ist, errechnet das low byte von
// cluster und setzt dann byteweise den inhalt:content.
// prüft ob buffer dirty (zu setztender cluster nicht in jetzt gepuffertem).
// prüfung erfolgt in fat_loadSector, dann wird alter vorher geschrieben, sonst gehen dort daten verloren !!
//***************************************************************************************************************
unsigned char fat_setCluster(unsigned long int cluster, unsigned long int content){  

	// FAT 16**************FAT 16	
	if(fat.fatType==16){					
													//	i=cluster/256
		unsigned long int i=cluster>>8;				// errechnet den sektor der fat in dem cluster ist (rundet immer ab) 
													// j=(cluster-256*i)*2 == 2*cluster-512*i
		unsigned long int j=(cluster<<1)-(i<<9);	// errechnet das low byte von cluster in dem ausgelesenen sektor der fat 
		
		if(0==fat_loadSector(i+fat.fatSec)){		//	neu laden (fat_loadSector prüft ob schon gepuffert)
//  			printf("\n\rfat_setClust: %lu-> %lu",cluster,content);
			bytesOfSec=&sector[j];					// init des zeigers auf low byte
			void *vc=&content;						// init des zeigers auf content
			*bytesOfSec++=*(unsigned char*)vc++;	// setzen von 2 byte..
			*bytesOfSec++=*(unsigned char*)vc++;
			bufferDirty=1;							// zeigt an, dass im aktuellen sector geschrieben wurde
			return (0); 											
			}	
		}

	// FAT 32**************FAT 32	
	else{
													 // i=cluster/128
		unsigned long int i=cluster>>7;				// errechnet den sektor der fat in dem cluster ist (rundet immer ab) 
													// j=(cluster-128*i)*4 == cluster*4-512*i
		unsigned long int j=(cluster<<2)-(i<<9);	// errechnet das low byte von cluster in dem ausgelesenen sektor der fat
				
		if(0==fat_loadSector(i+fat.fatSec)){		//	neu laden (fat_loadSector prüft ob schon gepuffert)			
//  			printf("\n\rfat_setClust: %lu-> %lu",cluster,content);
			bytesOfSec=&sector[j];					// init des zeigers auf low byte	
			void *vc=&content;						// init des zeigers auf content
			*bytesOfSec++=*(unsigned char*)vc++;	// setzen von 4 byte....
			*bytesOfSec++=*(unsigned char*)vc++;
			*bytesOfSec++=*(unsigned char*)vc++;
			*bytesOfSec++=*(unsigned char*)vc++;
			bufferDirty=1;							// zeigt an, dass im aktuellen sector geschrieben wurde
			return (0);
			}	
		}		


	return(1);									// neuladen des fat sektors, in dem oneCluster ist, nötig !!		
}
#endif


#if smallFileSys
//***************************************************************************************************************
// löscht cluster chain, beginnend ab dem startCluster.
// sucht cluster, setzt inhalt usw.. abschließend noch den cluster-chain ende markierten cluster löschen.
//***************************************************************************************************************
unsigned char fat_delClusterChain(unsigned long int startCluster){

  unsigned long int nextCluster=startCluster;		// tmp variable, wegen verketteter cluster..
	
  do{
	startCluster=nextCluster;
	nextCluster=fat_getNextCluster(startCluster);
	fat_setCluster(startCluster,0x00000000);  	
	}while(!((nextCluster==0xfffffff&&fat.fatType==32)||(nextCluster==0xffff&&fat.fatType==16)));	
	
	fat_writeSector(fat.currentSectorNr);

  return(0);
}
#endif


//***************************************************************************************************************
// Initialisiert die Fat(16/32) daten, wie: root directory sektor, daten sektor, fat sektor...
// siehe auch Fatgen103.pdf. ist NICHT auf performance optimiert!
// byte/sector, byte/cluster, anzahl der fats, sector/fat ... (halt alle wichtigen daten zum lesen ders datei systems!)
//*****************************************************************<**********************************************
unsigned char fat_loadFatData(unsigned long int sec){

 									// offset,size			
	unsigned int  		rootEntCnt;		// 17,2				// umbedingt mal nachsehen !
	unsigned int  		fatSz16;		// 22,2				sectors occupied by one fat16
	unsigned long int 	fatSz32;		// 36,4				sectors occupied by one fat32

	if(0==mmc_read_sector(sec,sector)){	// lesen von fat sector und bestimmen der wichtigen berreiche 
		
		fat.secPerClust=sector[13];			// fat.secPerClust, 13 only (power of 2)
		
		fat.bytsPerSec=sector[12];			// bytesPerSec, 12 high nibble 11 low nibble
		fat.bytsPerSec=fat.bytsPerSec<<8;
		fat.bytsPerSec+=sector[11];
		
		fat.fatSec=sector[15];				// resvdSecCnt, 15 high nibble 14 low nibble
		fat.fatSec=fat.fatSec<<8;
		fat.fatSec+=sector[14];

		rootEntCnt=sector[18];				// rootEntCnt, 18 high nibble 17 low nibble
		rootEntCnt=rootEntCnt<<8;
		rootEntCnt+=sector[17];
		
		fatSz16=sector[23];					// fatSz16, 23 high nibble 22 low nibble
		fatSz16=fatSz16<<8;
		fatSz16+=sector[22];
		
		
		fat.rootDir	 = (((rootEntCnt <<5) + fat.bytsPerSec) / fat.bytsPerSec)-1;		// ist 0 bei fat 32, sonst der root dir sektor	

		if(fat.rootDir==0){									// FAT32 spezifisch (die prüfung so, ist nicht spezifikation konform !).
		  void *vFatSz32=&fatSz32;								// void pointer auf fatSz32 zum schreiben von 4 bytes
		  *(unsigned char*)vFatSz32++=sector[36];				// lowest byte
		  *(unsigned char*)vFatSz32++=sector[37];				// 1. higher byte
		  *(unsigned char*)vFatSz32++=sector[38];				// 2. higher byte
		  *(unsigned char*)vFatSz32=sector[39];					// high byte

		  void *rootDir=&fat.rootDir;							// void pointer auf fat.rootDir zum schreiben von 4 bytes.
		  *(unsigned char*)rootDir++=sector[44];				// lowest byte
		  *(unsigned char*)rootDir++=sector[45];				// 1. higher byte
		  *(unsigned char*)rootDir++=sector[46];				// 2. higher byte
		  *(unsigned char*)rootDir=sector[47];					// high byte

		  fat.dataDirSec = fat.fatSec + (fatSz32 * sector[16]);	// data sector (beginnt mit cluster 2)		
		  fat.dir=0;											// dir auf 0==root dir, sonst 1. Cluster des dir											
		  fat.dataDirSec-=2;									// zeigt auf 1. cluster			
		  fat.fatType=32;										// fat typ					
		  fat.dataDirSec+=sec;									// addiert den startsektor auf (umrechnung von absolut auf real)
		  fat.fatSec+=sec;										// addiert den startsektor auf 	
		  }

		else{												// FAT16	spezifisch
			fat.dataDirSec = fat.fatSec + (fatSz16 * sector[16]) + fat.rootDir;		// data sektor (beginnt mit cluster 2)
			fat.rootDir=fat.dataDirSec-fat.rootDir;									// root dir sektor, da nicht im datenbereich(cluster)
			fat.dataDirSec-=2;														// zeigt auf 1. cluster		
			fat.dataDirSec+=sec;													// addiert den startsektor auf (umrechnung von absolut auf real)
			fat.fatSec+=sec;														// addiert den startsektor auf 	"
			fat.rootDir+=sec;														// addiert den startsektor auf 	"
			fat.dir=0;																// dir auf '0'==root dir, sonst 1.Cluster des dir
			fat.fatType=16;															// fat typ			
			}		

		return (0);
		}


	return (1);			// sector nicht gelesen, fat nicht initialisiert!!
}


//************************************************************************************************<<***************
// int fat sucht den 1. cluster des dateisystems (fat16/32) auch VBR genannt,  
// wenn superfloppy==0 wird der MBR ausgelesen um an VBR zu kommen.
//************************************************************************************************<<***************
unsigned char fat_initfat(void){					
  
  unsigned long int secOfFirstPartition=0;						// ist 1. sektor der 1. partition aus dem MBR  

	if(superfloppy==0){											// ist partitioniert
		if(0==mmc_read_sector(0,sector)){			

		  void *vSecOfFirstPartition=&secOfFirstPartition;		  

		  *(unsigned char*)vSecOfFirstPartition++=sector[454];
		  *(unsigned char*)vSecOfFirstPartition++=sector[455];
		  *(unsigned char*)vSecOfFirstPartition++=sector[456];
		  *(unsigned char*)vSecOfFirstPartition++=sector[457];

		  return fat_loadFatData(secOfFirstPartition);			// läd fat daten aus dem 1. sektor der patition
		  }

		}

	else {
		return fat_loadFatData(secOfFirstPartition);				// ist nicht partitioniert, läd fat daten aus sektor 0
		}

	return (1);
}
